"""
Computing LnZ of a graphical model using tensor networks
## Command line examples: 
* regular random graph with degree 3:
    python lnz.py -n50 -k 3 -fvsenum
    python lnz.py -n160 -k 3 -beta 0.9 -seed 3 -maxdim 34

* football c60 graph
    python lnz.py -beta 0.45 -seed 3 -maxdim 30 -verbose 1 -fvsenum -graph c60 # football graph
* 2d ferromagnetic Ising model
    python lnz.py -n20 -beta 0.45 -seed 3 -maxdim 30 -graph 2dsquare -verbose 1

python lnz.py -n10 -beta 0.45 -seed 3 -maxdim 32 -graph 2dsquare -verbose 1  -Dmax -1 -chi 100
python lnz.py -n10 -beta 0.45 -seed 3 -maxdim 32 -graph 2dsquare -verbose 1  -Dmax -1 -chi 100 -Jij randn -field randn

TODO:
    1. Optimize select_edge_total_dimension(). Storing total dimension as a variable, manitainig a sorted list for select edges.
"""
import torch
import numpy as np
import math
import networkx as nx
import string
import time
import sys
from tensor_network import Tensor_Network
from args import args
import random
def out_product(tensor,order):
    symmetirc_tensor=tensor
    for i in range(order-1):
     symmetirc_tensor = (symmetirc_tensor@tensor.transpose()).reshape(-1,1)
    return symmetirc_tensor     
def get_symmetric_tensor(order,bond):
    rank_number = order*bond
    weights = np.random.rand(rank_number)
    basis_vector = np.random.randn(rank_number,bond,1)
    symmetirc_tensor = weights[0]*out_product(basis_vector[0],order)
    for i in range(1,rank_number):
         symmetirc_tensor += weights[i]*out_product(basis_vector[i],order)
    symmetirc_tensor=symmetirc_tensor.reshape([bond]*order)
    return symmetirc_tensor  
def convert(filename,x):
    with open(filename,'r') as f:
         l=f.readlines()
    f.close
    n=ord(l[0][0])-63 #number of vertexs
    m=int(n*3/2) 
    p=int(n*(n-1)/2)
    last=p%6
    uptri_edges=[]
    for i in range(1,n,1):
       for j in range(0,i):
          uptri_edges.append([j,i])
    edges_index=[]
    line=l[x].strip()
    bitstring=''
    for j in range(1,len(line)):
       bitstring+=f'{ord(line[j])-63:6b}'
       #print(f'{ord(line[j])-63:6b}')
    for i in range(len(bitstring)):
        if bitstring[i]=='1':
           edges_index.append(i)
    
    
                

    return n,np.array(uptri_edges)[edges_index]    

if __name__ == '__main__':
    torch.autograd.set_detect_anomaly(True)

    #    torch.set_num_threads(8)

    device = torch.device("cpu" if args.cuda < 0 else "cuda:" + str(args.cuda))
    if args.graph == 'rrg' or args.graph == 'rer':
        graph = nx.random_regular_graph(args.k, args.n, seed=args.seed)
        edges = graph.edges
        print("regular random graph, n=", args.n, "k=", args.k, "seed=", args.seed, "maxdim=", args.maxdim)
    elif args.graph == 'gnp' or args.graph == 'ran' or args.graph == 'er':
        graph = nx.gnp_random_graph(args.n, 1.0 * args.c / args.n, seed=args.seed)
        edges = list(graph.edges)
        print("ER random graph, n=", args.n, "c=", args.c, "seed=", args.seed, "maxdim=", args.maxdim)
    elif args.graph == '2dsquare':
        graph = nx.grid_2d_graph(args.n, args.n, create_using=nx.Graph)
        graph = nx.Graph(graph)
        edges_2d = list(graph.edges)
        edges = [(i[0] * args.n + i[1], j[0] * args.n + j[1]) for i, j in edges_2d]
        args.L = args.n
        args.n = args.n ** 2
        print("2d lattice, L=", args.L, "seed=", args.seed, "maxdim=", args.maxdim)
    elif args.graph == 'tree':
        graph = nx.random_tree(args.n, seed=args.seed)
        edges = list(graph.edges)
        print(edges)
    elif args.graph == 'complete':
        graph = nx.complete_graph(args.n)
        edges = list(graph.edges)
    elif args.graph == "read_from_file":
        args.n,edges=convert('Generated_graphs.20.03.g6',args.which)
        print(args.n)
        args.graph='rrg'
        #print(edges)
        print(len(edges))
     
    
    edges = np.unique(np.array([sorted(a) for a in edges]), axis=0)
    
    if args.permutation:
       randomindex=[i for i in range(args.n)]
       random.shuffle(randomindex)
       randomindex=np.array(randomindex)
       edges[:,0]=randomindex[edges[:,0]]
       edges[:,1]=randomindex[edges[:,1]]
    
    G = nx.Graph()
    G.add_nodes_from(np.arange(args.n))
    G.add_edges_from(edges)
    if args.wl:
      
      n2,edges2=convert('Generated_graphs.20.03.g6',args.which+1)
      G1=nx.Graph()
      G1.add_nodes_from(np.arange(n2))
      G1.add_edges_from(edges2)
      from wlkernel.weisfeiler_lehman import is_possibly_isomorphic
      time_1=time.time()
      print(is_possibly_isomorphic(G,G1))
      print("time: {:.2g} sec".format(time.time()-time_1))
    if  args.nx_iso:
     if args.graph=='read_from_file':
      n2,edges2=convert('Generated_graphs.42.05.g6',args.which+20)
      G1=nx.Graph()
      G1.add_nodes_from(np.arange(n2))
      G1.add_edges_from(edges2)
     else:
      G1=nx.random_regular_graph(args.k,args.n)
     time_1=time.time()
     print(nx.is_isomorphic(G,G1))
     print("time: {:.2g} sec".format(time.time()-time_1))
    
    G_backup = G.copy()
    t0 = time.time()
    if args.seed2 < 0:
        args.seed2 = args.seed

    beta = torch.tensor([args.beta], dtype=torch.float64, device=device, requires_grad=True)
    max_degree = max(np.array(G.degree)[:, 1])
    np.random.seed(args.seed1)
    torch.manual_seed(args.seed1)
    #tensor1=[]
    tensor1=torch.nn.ParameterList([])
    tensor2=torch.randn(args.bond,args.bond,dtype=torch.float64,device=device)
    tensor2=torch.triu(tensor2)+torch.triu(tensor2,diagonal=1).t()
    y=torch.exp(tensor2)
    eig,v=torch.symeig(y)
    print(eig)
    print(v)
    print(tensor2)
    for i in range(1,max_degree+1):
        tensor1.append(torch.nn.Parameter(torch.from_numpy(get_symmetric_tensor(i,args.bond))))
    print(tensor1)
    '''
    if args.raw:
        args.node = "raw"
    '''
    np.savetxt('edges_{}.txt'.format(args.which),edges)
    tn = Tensor_Network(args.n, edges, args.choice, args.bond,tensor1,tensor2, beta, seed=args.seed2, maxdim=args.maxdim,
                        verbose=args.verbose, Dmax=args.Dmax, chi=args.chi, node_type=args.node)
    # tn.tensors[0].tensor.norm().backward()
    t0 = time.time()
    lnZ_tn = tn.contraction()
    (lnZ_tn).backward()
    lnZ_tn = lnZ_tn
    print("lnZ_tn = {:.15g}, time: {:.2g} Sec. maxdim_inter={:d}".format(lnZ_tn.item(), time.time() - t0,
                                                                         int(tn.maxdim_intermediate)))
    print("beta.grad={:.15g}".format(beta.grad.item()))
    print(tensor1)
    print(tensor1[2].grad)
    if args.output:
       with open('{}_n={}_seed1_{}_beta_{}.txt'.format(args.graph,args.n,args.seed1,args.beta),'a') as fp:
          fp.write('{}  {:.15g} {:.15g}\n'.format(args.which,lnZ_tn.item(),beta.grad.item()))
       fp.close()
    '''
    correlation_tn = W.grad
    print('entropy: ', (-beta ** 2 * beta.grad).item())
    print('energy: ', (-(lnZ_tn * tn.n) / beta - beta * beta.grad).item())
    print(correlation_tn)
    print(h.grad)
    print(edges)
    '''


